import { d3 } from '../d3-bundle';
import { SeriesType } from '../types/types';
import { ChartDataService } from './chart.data.service';
import { COLOR_WHEEL } from './color-wheel';
export class ChartD3Service {
    static addArea(svg, line, y, data, color) {
        const area = d3.area()
            .curve(d3.curveMonotoneX)
            .defined(line.defined())
            .x(line.x())
            .y1(line.y())
            .y0(y(0));
        svg.append('path')
            .data([data])
            .attr('class', 'series area')
            .attr('fill', color)
            .attr('d', area);
        return area;
    }
    static addStackedColumn(svg, x, y, data, color) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g')
            .attr('class', 'group')
            .style('fill', color);
        const chartBars = seriesContainers
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d, i) => {
            return d[0] < 0 ? `bar negative bar-${i}` : `bar positive bar-${i}`;
        })
            .attr('width', x.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.data.categoryValue) + (ChartD3Service.groupPadding / 2))
            .attr('y', y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => y(d[1]))
            .attr('height', (d) => y(d[0]) - y(d[1]));
        return chartBars;
    }
    static addColumn(svg, x, x1, y, data, color, _dims, index) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g');
        const chartBars = seriesContainers
            .attr('class', 'series')
            .attr('transform', 'translate(' + x1(index) + ',0)')
            .style('fill', color)
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d) => {
            return d.y < 0 ? 'bar negative' : 'bar positive';
        })
            .style('opacity', (d) => {
            return d.y < 0 ? .6 : 1;
        })
            .attr('width', x1.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.x) + (ChartD3Service.groupPadding / 2))
            .attr('y', () => y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => d.y >= 0 ? (y(d.y)) : y(0))
            .attr('height', (d) => {
            return Math.max(Math.abs(y(d.y) - y(0)), 1);
        });
        return chartBars;
    }
    static addPie(svg, data, dims) {
        const oRadius = dims.height / 2 - dims.margin.top - dims.margin.bottom;
        // construct default pie layout
        const pie = d3.pie()
            .padAngle(.02)
            .value((d) => d.y)
            .sort(null);
        // construct arc generator
        const arc = d3.arc()
            .outerRadius(oRadius)
            .innerRadius(1);
        // creates the pie chart container
        const g = svg.append('g')
            .attr('transform', 'translate(' + (dims.width / 2) + ',' + (dims.height / 2) + ')');
        // add header
        const header = svg.append('g')
            .attr('transform', 'translate(10, 20)')
            .attr('class', `header`);
        // enter data and draw pie chart
        const slices = ChartD3Service.renderSlices(g, pie, arc, data);
        return { slices, g, pie, arc, header };
    }
    static addLine(svg, x, y, data, color, secondary = false) {
        const lineClass = secondary ? 'anj-line anj-secondary' : 'anj-line';
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        const valueLine = d3.line()
            .curve(d3.curveMonotoneX)
            .defined((d) => d.y !== null)
            .x((d) => x(d.x) + xOffset)
            .y((d) => y(d.y));
        svg.append('path')
            .data([data])
            .attr('class', lineClass)
            .attr('stroke', color)
            .attr('d', valueLine)
            .attr('stroke-dasharray', function () {
            return this.getTotalLength();
        })
            .attr('stroke-dashoffset', function () {
            return this.getTotalLength();
        })
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('stroke-dashoffset', 0);
        return valueLine;
    }
    static addOverlay(svg, dims) {
        const overlay = svg.append('rect')
            .attr('class', 'anj-overlay')
            .attr('width', dims.width)
            .attr('height', dims.height);
        return overlay;
    }
    static addPoints(svg, x, y, data, color, index, options) {
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        options = options || {
            opacity: 0,
            fill: undefined,
            radius: 1
        };
        return svg.selectAll(`.point.series${index}`)
            .data(data)
            .enter()
            .append('svg:circle')
            .style('opacity', options.opacity)
            .style('stroke', color)
            .style('fill', options.fill)
            .attr('class', (_d, i) => `point point-${i}`)
            .attr('r', () => options.radius)
            .attr('cx', (d) => x(d.x) + xOffset)
            .attr('cy', (d) => y(d.y));
    }
    static addRevealAnimation(svg, dims, axisConfigs, data) {
        const delay = ChartD3Service.focusLength;
        const duration = ChartD3Service.animationLength;
        const curtainHeight = dims.height;
        const curtainWidth = dims.width;
        const leftPad = this.calculateLeftPad(dims, axisConfigs, data);
        const curtain = svg.append('rect')
            .attr('x', 0 + leftPad)
            .attr('y', 0)
            .attr('height', curtainHeight)
            .attr('width', curtainWidth - leftPad + dims.margin.right)
            .attr('class', 'curtain')
            .style('fill', '#ffffff');
        const t = svg.transition()
            .delay(delay)
            .duration(duration);
        t.select('rect.curtain')
            .attr('x', curtainWidth + dims.margin.right)
            .attr('width', 0);
        return curtain;
    }
    static addStackedScrubbers(svg, x, data, dims, index) {
        const xOffset = ((x.bandwidth() + ChartD3Service.calculatePadding(data, dims)) / 2);
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y1', 0)
            .attr('x2', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`)
            .attr('stroke', '#ffffff')
            .style('opacity', '0');
    }
    static addScrubbers(svg, x, data, dims, index) {
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.x))
            .attr('y1', 0)
            .attr('x2', (d) => x(d.x))
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`);
    }
    static getX1(x, data, multiSeries) {
        const seriesCount = (multiSeries) ? data.length : 1;
        return d3.scaleBand()
            .rangeRound([0, x.bandwidth()])
            // Sets cluster count
            .domain(d3.range(seriesCount))
            .padding(0);
    }
    static addXAxis(svg, dims, data, axisConfigs, _minMax, primarySeries, y) {
        const x = ChartD3Service.axisTyper(primarySeries, data, dims)
            // Defines physical width of the axis
            .range([this.calculateLeftPad(dims, axisConfigs, data), dims.width]);
        // Ticks get crowded when the chart resizes too small. This is our calculation to handle that for non-Line Charts
        const uniqueX = Array.from(new Set(data.map(item => item.x.toString())));
        const availableSlots = (dims.width / 100);
        const tickModulo = Math.max(Math.floor(uniqueX.length / availableSlots), 1);
        const xAxis = (primarySeries === SeriesType.Line || primarySeries === SeriesType.Scatter)
            // Line Charts need to have the ticks limited by svg, see below
            ? d3.axisBottom(x)
            // Other Charts can have d3 handle the tick limiting up front
            : d3.axisBottom(x).tickValues(x.domain().filter((_d, i) => !(i % tickModulo)));
        // Defines size of outside start/stop ticks. In this case, no ticks
        xAxis.tickSizeOuter(0)
            // Defines formatting of tick text
            .tickFormat(ChartDataService.xFormatter.bind(this, ChartDataService.calculateInterval(data), axisConfigs.x))
            .scale(x);
        // Draw the axis to the svg
        svg.append('g')
            .attr('class', 'x axis')
            // Adjust the axis up/down
            .attr('transform', `translate(0,${dims.height})`)
            .call(xAxis);
        // Align x-axis path with 0 on the y-axis
        svg.select('.x.axis path')
            .attr('transform', 'translate(0,' + (y(0) - dims.height) + ')');
        if (!axisConfigs.x.showAxis) {
            svg.select('.x.axis').style('display', 'none');
        }
        if (!axisConfigs.x.showTicks) {
            svg.selectAll('.x.axis .tick').style('display', 'none');
        }
        // Handle tick spacing for LineCharts
        if (primarySeries === SeriesType.Line || primarySeries === SeriesType.Scatter) {
            const ticksLength = svg.selectAll('.x.axis .tick').size();
            const slots = (dims.width / 100);
            const modulo = Math.max(Math.floor(ticksLength / slots), 1);
            svg.selectAll('.x.axis .tick').filter((_d, i) => (i % modulo)).style('display', 'none');
        }
        return x;
    }
    static axisTyper(type, data, dims) {
        switch (type) {
            case SeriesType.Line:
                return d3.scaleTime()
                    // Defines range of values displayed (start/stop)
                    .domain(ChartDataService.determineDateRange(data));
            case SeriesType.Scatter:
                const xArray = data.map((d) => d.x);
                return d3.scaleLinear()
                    // Defines range of values displayed (start/stop)
                    .domain([Math.min(...xArray), Math.max(...xArray)]);
            default:
                const domain = data.map((d) => d.x.toString());
                return d3.scaleBand()
                    .padding(ChartD3Service.calculatePadding(data, dims))
                    // Defines range of values displayed (start/stop)
                    .domain(domain);
        }
    }
    static calculatePadding(data, dims) {
        const colWidth = dims.width / (data.length - 1);
        return colWidth > 30 ? .05 : 0;
    }
    static addYAxis(svg, dims, data, axisConfigs, minMax) {
        // minMax can override normal domain calculation, allowing for negative values
        const domain = (minMax) ? [minMax.min, minMax.max] : [Math.min(d3.min(data, (d) => d.y), 0), d3.max(data, (d) => d.y)];
        const y = d3.scaleLinear()
            // Defines range of values displayed (start/stop)
            .domain(domain)
            // Ensures a first and last tick on the axis
            .nice()
            // Defines physical height of the axis
            .range([dims.height, dims.margin.top]);
        const yAxis = d3.axisLeft(y)
            // Defines size of outside start/stop ticks. In this case, no ticks
            .tickSize(0)
            // Defines number of ticks on axis
            .ticks(this.calculateTicks(dims.height))
            // Defines formatting of tick text
            .tickFormat(ChartDataService.yFormatter.bind(this, axisConfigs.y))
            .scale(y);
        // Draw the axis to the svg
        svg.append('g')
            .attr('class', 'y axis')
            // Adjust the axis left/right
            .attr('transform', `translate(${this.calculateLeftPad(dims, axisConfigs, data)},0)`)
            .call(yAxis);
        if (!axisConfigs.y.showAxis) {
            svg.select('.y.axis').style('display', 'none');
        }
        if (!axisConfigs.y.showTicks) {
            svg.selectAll('.y.axis .tick').style('display', 'none');
        }
        return y;
    }
    static createSVG(dims) {
        const svg = d3.create('svg');
        svg
            .attr('width', dims.width + dims.margin.left)
            .attr('height', dims.height + dims.margin.top)
            .attr('transform', `translate(${dims.margin.left}, ${dims.margin.top})`);
        return svg;
    }
    static getIndexOfClosestPoint(x, data) {
        // Capture granular, generated point - does not correspond to a given data point
        const x0 = x.invert(d3.mouse(d3.event.currentTarget)[0]);
        const bisector = d3.bisector((d) => d.x).left;
        const index = bisector(data, x0, 1);
        const d0 = data[index - 1];
        const d1 = data[index];
        const rangeValueOfFirst = x((d0) ? d0.x : null), rangeValueOfSecond = x((d1) ? d1.x : null), rangeValueOfMousePos = d3.mouse(d3.event.currentTarget)[0], 
        // closestXValue = Math.abs(rangeValueOfMousePos - rangeValueOfFirst) > Math.abs(rangeValueOfMousePos - rangeValueOfSecond) ? d1 : d0,
        closestIndex = Math.abs(rangeValueOfMousePos - rangeValueOfFirst) > Math.abs(rangeValueOfMousePos - rangeValueOfSecond) ? index : index - 1;
        return closestIndex;
    }
    static hideFocus(svg) {
        svg.selectAll('.point')
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 1)
            .style('opacity', '0');
        svg.selectAll(`.scrubber`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '0');
    }
    static setPieNav(header, parentName, childName) {
        header.selectAll('text').remove();
        if (childName) {
            const drillup = header.append('text')
                .text(parentName)
                .attr('class', 'drill-up');
            header.append('text')
                .attr('x', header.select('text').node().getBoundingClientRect().width + 5)
                .text(` - ${childName}`);
            return drillup;
        }
        else {
            header.append('text')
                .text(parentName);
        }
    }
    static showFocus(svg, activeIndex) {
        ChartD3Service.hideFocus(svg);
        svg.selectAll(`.point-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 4)
            .style('opacity', '1');
        svg.selectAll(`.scrubber-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '1');
    }
    static renderSlices(g, pie, arc, data) {
        let _current;
        g.datum(data).selectAll('path')
            .data(pie)
            .enter().append('path')
            .attr('class', 'piechart');
        g.datum(data)
            .selectAll('path')
            .data(pie)
            .attr('fill', (_d, i) => _d.data.color || COLOR_WHEEL[i])
            .style('cursor', (_d) => _d.data.drilldown ? 'pointer' : 'inherit')
            .transition().duration(1000).attrTween('d', (a) => {
            const i = d3.interpolate(_current, a);
            _current = i(0);
            return (t) => {
                return arc(i(t));
            };
        });
        g.datum(data).selectAll('path')
            .data(pie).exit().remove();
        return g.selectAll('path');
    }
    static calculateLeftPad(dims, axisConfigs, data) {
        // Make sure the y gutter is wide enough to support our longest values
        const maxYValueLength = Math.max(...data.map(d => d.y)).toString().length;
        this.yPad = maxYValueLength * 4;
        if (axisConfigs.y.showAxis || axisConfigs.y.showTicks) {
            return this.yPad + dims.margin.left;
        }
        else {
            return 0;
        }
    }
    static calculateTicks(dim) {
        return Math.max(dim / 100, 2);
    }
    static formatDataAsStacked(rawData) {
        const stackFactory = d3.stack()
            .offset(ChartD3Service.stackOffsetDiverging)
            // TODO: get proper name from somewhere
            .keys(rawData.map((_s, i) => `namey-${i}`));
        const data = ChartD3Service.seriesAsStackFormat(rawData);
        return stackFactory(data);
    }
    static seriesAsStackFormat(series) {
        // TODO: get proper name from somewhere
        const namedSeries = series.map((s, i) => {
            return {
                name: `namey-${i}`,
                coords: s
            };
        });
        const collector = [];
        namedSeries.forEach((s) => {
            s.coords.forEach((coords, i) => {
                if (!collector[i]) {
                    collector[i] = {};
                }
                collector[i].categoryValue = coords.x;
                collector[i][s.name] = coords.y;
            });
        });
        return collector;
    }
    static stackOffsetDiverging(series, seriesOrder) {
        const n = series.length;
        if (n < 1) {
            return;
        }
        const firstSeriesIndex = seriesOrder[0];
        /* tslint:disable */
        for (let i, j = 0, dy, yp, yn, columnCount = series[firstSeriesIndex].length; j < columnCount; ++j) {
            for (yp = yn = 0, i = 0; i < n; ++i) {
                const serie = series[seriesOrder[i]];
                const dataPoint = serie[j];
                dy = dataPoint[1] - dataPoint[0];
                if (dy >= 0) {
                    dataPoint[0] = yp;
                    dataPoint[1] = yp += dy;
                }
                else if (dy < 0) {
                    dataPoint[1] = yn;
                    dataPoint[0] = yn += dy;
                }
                else {
                    dataPoint[0] = yp;
                }
            }
        }
    }
}
ChartD3Service.animationLength = 2500;
ChartD3Service.focusLength = 400;
ChartD3Service.yPad = 15;
ChartD3Service.groupPadding = 2;
